using Godot;
using System;
using System.Collections;
using System.Collections.Generic;

public class TileGrid : Resource, IReadOnlyList<Tile>
{
    [Export]
    public int Size { get; private set; } = 10;

    [Export]
    private Resource _startTileTypeResource;
    private TileType StartTileType => (TileType)_startTileTypeResource;

    public int Count => Size * Size;

    public Tile this[int index]
    {
        get => Tiles[index];
        set => Tiles[index] = value;
    }

    public Tile this[int x, int y]
    {
        get
        {
            MapPosition(x, y, out int idx);
            return Tiles[idx];
        }
        set
        {
            MapPosition(x, y, out int idx);
            Tiles[idx] = value;
        }
    }

    private List<Tile> Tiles
    {
        get
        {
            if (_tiles == null)
            {
                _tiles = new List<Tile>();
                Generate(_tiles);
            }

            return _tiles;
        }
    }
    private List<Tile> _tiles = null;

    private void Generate(List<Tile> tiles)
    {
        tiles.Clear();
        for (int i = 0; i < Count; i++)
        {
            var tile = new Tile();
            tile.isHighlighted = false;
            tile.temperature = 0.0f;
            tile.type = StartTileType;
            tile.currentHealth = 1.0f;

            tiles.Add(tile);
        }
    }

    public IEnumerator<Tile> GetEnumerator()
    {
        return Tiles.GetEnumerator();
    }

    IEnumerator IEnumerable.GetEnumerator()
    {
        return Tiles.GetEnumerator();
    }

    public void MapIndex(int idx, out int x, out int y)
    {
        y = idx / Size;
        x = idx % Size;
    }

    public void MapPosition(int x, int y, out int idx)
    {
        idx = y * Size + x;
    }

    public bool IsInBounds(int x, int y)
    {
        return x >= 0 && x < Size && y >= 0 && y < Size;
    }

    private readonly int[] _neighbours = new int[]
    {
        -1, 0,
        0, 1,
        1, 0,
        0, -1
    };

    private void GetCoordinate(int dir, out int x, out int y)
    {
        dir %= 4;
        int idx = dir * 2;
        x = _neighbours[idx];
        y = _neighbours[idx + 1];
    }

    public float GetNeighbourProportion<T>(int x, int y)
    {
        int typed = GetNeighbourCount<T>(x, y);
        int all = GetNeighbourCount(x, y);
        return (float)typed / (float)all;
    }

    public int GetNeighbourCount<T>(int x, int y)
    {
        int count = 0;
        for (int i = 0; i < 4; i++)
        {
            GetCoordinate(i, out var nx, out var ny);

            if (!IsInBounds(x + nx, y + ny))
                continue;

            if (!(this[x + nx, y + ny].type is T))
                continue;

            count++;
        }

        return count;
    }

    public int GetNeighbourCount(int x, int y)
    {
        int count = 0;

        for (int i = 0; i < 4; i++)
        {
            GetCoordinate(i, out var nx, out var ny);

            if (!IsInBounds(x + nx, y + ny))
                continue;

            count++;
        }

        return count;
    }
}

